import logging
import sys
import unittest

from lbmessaging.services.Common import BaseService, ExposedIn

mapping1 = {
        'send': 'dumb_send',
        'receive': 'dumb_receive'
    }

mapping2 = {
        'send_2': 'dumb_send',
        'receive_2': 'dumb_receive'
    }

@ExposedIn("DumbService", mapping=mapping1)
@ExposedIn("DumbService2", mapping=mapping2)
class DumbExchange():

    def __init__(self, channel):
        self.channel = channel
        self.lastAction = None

    def dumb_send(self, arg1, kwarg1=None):
        print(arg1),
        print(kwarg1)
        self.lastAction = 'SENT'

    def dumb_receive(self):
        self.lastAction = 'RECEIVED'
        return 'RECEIVED'


@ExposedIn("DumbService")
class DumbExchange2():
    pass


class DumbService(BaseService):
    pass


class DumbService2(BaseService):
    pass


class TestExposeIn(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_meta_usage(self):
        srv = DumbService()
        try:
            srv.send()
            self.fail("Send with no arguments")
        except:
            pass
        srv.send('arg1', kwarg1='toto')
        res = srv.receive()
        self.assertEqual(res, 'RECEIVED')

    def test_meta_setup(self):
        srv = DumbService()
        for funct in mapping1.keys():
            self.assertTrue(funct in dir(srv))
        for funct in mapping2.keys():
            self.assertFalse(funct in dir(srv))
        srv = DumbService2()
        for funct in mapping2.keys():
            self.assertTrue(funct in dir(srv))

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



###############################################################################
# (c) Copyright 2017 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Module grouping the common build functions.
'''

import datetime
import json
import logging
import uuid
from collections import namedtuple

import os
import pika
import uuid

__author__ = 'Ben Couturier <ben.couturier@cern.ch>'


# Utility class to ease the use of the messages
Message = namedtuple('Message', ['body', 'method_frame', 'header_frame'],
                     rename=True)


# Our own exception to signal that the processing was interrupted and
# we shouldn't ack the message
class InterruptedProcessing(Exception):
    pass


# Various utilities for connection management
def _get_pwd_from_sys():
    """
    Get the RabbitMQ password from the environment of from a file on disk
    :returns a username, password tuple for RabbitMQ connection
    """
    # First checking the environment
    res = os.environ.get("RMQPWD", None)

    # Checking for the password in $HOME/private/rabbitmq.txt
    if res is None:
        fname = os.path.join(os.environ["HOME"], "private", "rabbitmq.txt")
        if os.path.exists(fname):
            with open(fname, "r") as f:
                data = f.readlines()
                if len(data) > 0:
                    res = data[0].strip()
    if res is None:
        # At this point we should have found the authentication
        # Otherwise, we throw an exception...
        raise Exception("Could not locate broker credentials")


    # Separate the username/password
    (username, password) = res.split("/")
    return username, password


def get_connection(host=None,
                   username=None, passwd=None,
                   port=5671, vhost='/lhcb', use_ssl=True,
                   heartbeat_interval=None,
                   blocked_connection_timeout=None):
    """
    Create a connection to RabbitMQ
    :param host: the RabbitMQ server hostname
    :param username: the RabbitMQ server username
    :param passwd: the RabbitMQ server password
    :param port: the RabbitMQ server port
    :param vhost: the RabbitMQ virtual host that should be used in the
                  connection
    :param use_ssl: if True, an encrypted connection is created, plain text
                    otherwise
    :returns: a pika BlockingConnection to RabbitMQ generated with the
              arguments
    """

    envhost = os.environ.get("RMQHOST", None)
    if envhost:
        host = envhost
    if host is None:
        host = "lbmessagingbroker.cern.ch"

    envvhost = os.environ.get("RMQVHOST", None)
    if envvhost:
        vhost = envvhost

    if os.environ.get("RMQNOSSL", None):
        use_ssl = False

    if os.environ.get("RMQPORT", None):
        port = int(os.environ.get("RMQPORT"))

    if username is None or passwd is None:
        (username, passwd) = _get_pwd_from_sys()
    credentials = pika.PlainCredentials(username, passwd)
    cpar = " ".join([str(x) for x in [host, use_ssl, port, vhost]])
    logging.debug("Connection paremeters: %s" % cpar)
    conn_params = { "ssl":use_ssl,
                    "port":port,
                    "virtual_host":vhost,
                    "credentials":credentials }
    if heartbeat_interval is not None:
        conn_params["heartbeat_interval"] = heartbeat_interval
    if blocked_connection_timeout is not None:
        conn_params["blocked_connection_timeout"] = blocked_connection_timeout

    params = pika.ConnectionParameters(host, **conn_params )
    return pika.BlockingConnection(params)


def check_channel(channel, connection=None):
    """
    Ugly hack to keep backwards compatibility
    We should receive a pika.Channel, but if we are given a
    pika.BlockingConnection, we just create a new channel from it....

    :param channel: a pikaBlockingConnection channel to verify
    :param connection: a pikaBlockingConnection from which the channel needs
                       to be extracted
    :returns: the correct channel that should be used in the transactions
    """
    checked_channel = None
    if connection is not None:
        checked_channel = connection.channel()
    else:
        if isinstance(channel, pika.BlockingConnection):
            checked_channel = channel.channel()
        else:
            checked_channel = channel
    return checked_channel


class QueueHelper(object):
    """
    Utility class to access the rabbitmq broker
    """

    def __init__(self, channel=None, connection=None):
        """
        Initialize the messenger class
        :param channel: a pikaBlockingConnection channel that should be used in
                        the broker
        :param connection: a pikaBlockingConnection that should be used in
                        the broker
        """

        self._channel = check_channel(channel, connection)
        import lbmessaging
        self.MAX_PRIORITY = lbmessaging.priority(lbmessaging.EMERGENCY)

    def _setup_exchange(self, channel, exchange_name, auto_delete=False):
        """
        Creates an exchange with the desired exchange_type
        :param channel: a pikaBlockingConnection channel that should be used in
                        the exchange declarations
        :param exchange_name: the exchange name used in the broker
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        """
        channel.exchange_declare(exchange=exchange_name,
                                 exchange_type='topic',
                                 auto_delete=auto_delete,
                                 durable=True)

    def basic_publish(self, exchange_name, routing_key, body, priority=1,
                      max_retry=3, headers=None, auto_delete=False,
                      expiration=None, extra_properties={}):
        """
        Send a message to the topic defined for the builds
        :param exchange_name: the exchange that should be used
        :param routing_key: the routing (binding) key used to route the message
                            from the exchange to the queue that is bind with
                            this key to the exchange
        :param body: the payload of the message
        :param priority: the priority of the message (0 - 255)
        :param max_retry: the number of retries of sending the message in case
                          of failure
        :param headers: the headers that are included in the message
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        :param expiration: TTL for the message. The message will be
                           automatically deleted if this value is reached.
                           The value should be in milliseconds. For non
                           expiring messages, the flag should be set to None
        :param extra_properties: Extra pika BasicProperties to be set.
        """
        channel = self._channel
        self._setup_exchange(channel, exchange_name,
                             auto_delete=auto_delete)
        now = datetime.datetime.utcnow().isoformat()
        ipd = now
        if headers and headers.get('initial_publish_date'):
            ipd = headers.get('initial_publish_date')
        uid = uuid.uuid4().hex
        if headers and headers.get('uid'):
            uid = headers.get('uid')
        props = pika.BasicProperties(delivery_mode=2, priority=priority,
                                     headers={'max_retry': max_retry,
                                              'initial_publish_date': ipd,
                                              'uid': uid,
                                              # make message persistent
                                              'publish_date': now})
        if expiration:
            # It must be a string in milliseconds
            props.expiration = str(expiration)
        for key, value in extra_properties.items():
            setattr(props, key, value)
        channel.basic_publish(exchange=exchange_name,
                              routing_key=routing_key,
                              body=body,
                              properties=props)

    def bind_queue_to_exchange(self, queue_name, exchange_name,
                               routing_keys=None,
                               route_only_to_queue=False,
                               auto_delete=False):
        """
        Binds the queue to an exchange, make sure the queue is persistent
        :param queue_name: the queue on which needs to be bind to the exchange
        :param exchange_name: the exchange to which the queue should be bind to
        :param routing_keys: the routing keys that are used to route a message
                             from the exchange to the queue
        :param route_only_to_queue: if true and no routing keys are defined,
                                    the queue will listen only for the messages
                                    destined for the queue, i.e. the routing
                                    keys in the messages are identical with the
                                    queue name.
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        """

        return self._setup_queue(self._channel, exchange_name,
                                 queue_name, routing_keys,
                                 route_only_to_queue=route_only_to_queue,
                                 auto_delete=auto_delete)

    def _setup_queue(self, channel, exchange_name, queue_name=None,
                     routing_keys=None, route_only_to_queue=False,
                     auto_delete=False):
        """
        Setup a queue and binds
        :param channel: a pikaBlockingConnection channel that should be used in
                        the broker
        :param queue_name: the queue on which needs to be bind to the exchange
        :param exchange_name: the exchange to which the queue should be bind to
        :param routing_keys: the routing keys that are used to route a message
                             from the exchange to the queue
        :param route_only_to_queue: if true and no routing keys are defined,
                                    the queue will listen only for the messages
                                    destined for the queue, i.e. the routing
                                    keys in the messages are identical with the
                                    queue name.
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        :returns: the queue name. If in the args, the queue name was none,
                  an anonymous queue is created and its name is returned.
        """
        self._setup_exchange(channel, exchange_name,
                             auto_delete=auto_delete)
        if queue_name is None:
            # Anonymous queue is NOT persistent
            logging.info("Declaring anonymous queue")
            result = channel.queue_declare(exclusive=True,
                                           auto_delete=auto_delete,
                                           arguments={
                                               "x-max-priority":
                                                   self.MAX_PRIORITY})
            queue_name = result.method.queue
        else:
            # Named queues are persistent...
            logging.info("Declaring durable queue: %s" % queue_name)
            result = channel.queue_declare(durable=1, queue=queue_name,
                                           auto_delete=auto_delete,
                                           arguments={
                                               "x-max-priority":
                                                   self.MAX_PRIORITY})

        # defaults to the wildcard for topic channels
        if routing_keys is None:
            # If the routing keys list is None, we could listen for all the
            # messages in the exchange using # or we could listen only for the
            # messages for this specific queue. As an example, in the RPC mode,
            # both the client and the server are using the same exchange, but
            # we need to separate the messages for the server queue ( requests)
            # from those for the client anonymous queue (replays)
            if route_only_to_queue:
                routing_keys = [queue_name]
            else:
                routing_keys = ["#"]

        # Now binding the queue to the topic
        for routing_key in routing_keys:
            logging.info("Binding %s to %s for '%s'" % (queue_name,
                                                        exchange_name,
                                                        routing_key))
            channel.queue_bind(exchange=exchange_name,
                               queue=queue_name,
                               routing_key=routing_key)
        return queue_name


class Sender(QueueHelper):
    """
    Class used by cvmfs-lhcbdev.cern.ch to communicate with the outside world
    """

    def __init__(self, exchange, *args, **kwargs):
        """
        Initialize props
        :param exchange_name: the exchange name used in the broker
        """
        super(Sender, self).__init__(*args, **kwargs)
        self._exchange_name = exchange

    def send_message(self, routing_key, payload, priority=1, max_retry=3,
                     auto_delete=False, expiration=None):
        """
        Sends a message to the exchange
        :param routing_keys: the routing keys set in the message in order to
                             route it from the exchange to the queue(s)
        :param payload: the content of the messages send to the exchange
        :param priority: the priority of the message
        :param max_retry: the max of retries in case of command failure on CVMFS
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        :param expiration: the TTL of the message
        """
        body = json.dumps(payload)
        self.basic_publish(self._exchange_name, routing_key, body,
                           priority=priority, max_retry=max_retry,
                           expiration=expiration,
                           auto_delete=auto_delete
                           )
        logging.info("Publish done to %s" % (self._exchange_name))

    def resend_message(self, message, keep_retry_count=False):
        """
        Resend a message based on a received header frame payload
        :param message: the message whose consumption failed as Message object
        :param keep_retry_count: if true, the retry count is not decremented.
                                 Otherwise, the retry count is decremented by 1
        """
        routing_key = message.method_frame.routing_key

        bodyjson = json.dumps(message.body)
        new_max_retry = message.header_frame.headers.get('max_retry', 0)
        if not keep_retry_count and 'max_retry' in message.header_frame.\
                headers.keys():
            new_max_retry = message.header_frame.headers['max_retry'] - 1

        self.basic_publish(self._exchange_name, routing_key, bodyjson,
                           priority=message.header_frame.priority,
                           max_retry=new_max_retry,
                           headers=message.header_frame.headers)
        logging.info("resend done to %s" % (self._exchange_name))


class Receiver(QueueHelper):
    """
    Initializes a queue to receive messages
    """

    def __init__(self, exchange, queue, routing_keys=None, auto_delete=False,
                 route_only_to_queue=False, no_bind=False,
                 *args, **kwargs):
        """
        Initialize props
        :param exchange_name: the exchange name used in the broker
        :param queue: the queue on which needs to be bind to the exchange
        :param routing_keys: the routing keys that are used to route a message
                             from the exchange to the queue
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        :param route_only_to_queue: if true and no routing keys are defined,
                                    the queue will listen only for the messages
                                    destined for the queue, i.e. the routing
                                    keys in the messages are identical with the
                                    queue name.
        """
        super(Receiver, self).__init__(*args, **kwargs)
        self._exchange_name = exchange
        self._queue_name = queue
        if not no_bind:
            self.bind_queue_to_exchange(
                queue, exchange_name=exchange,
                routing_keys=routing_keys,
                route_only_to_queue=route_only_to_queue,
                auto_delete=auto_delete)

    def receive_messages(self):
        """
        Receive all messages in the queue and returns the list
        :returns a list of Message objects representing
                 all the messages in the queue
        """
        message_list = []
        while True:
            message = self.receive_message()
            if message is None:
                break
            message_list.append(message)
        return message_list

    def receive_message(self, runBeforeAck=None, noAck=False):
        """
        Receive a message from the queue
        :param runBeforeAck: A method to be invoked before sending the ACK back
                             to the message broker
        :param noAck: if True, no ACK is sent back to the message broker
        :return: the message itself as Message object
        """
        message = None
        channel = self._channel
        method_frame, header_frame, body = channel.basic_get(
            queue=self._queue_name)
        if method_frame is not None:
            try:
                t = json.loads(body.decode('UTF-8'))
            except:
                t = body
            message = Message(t, method_frame, header_frame)
            if runBeforeAck:
                runBeforeAck(message)
            if not noAck:
                channel.basic_ack(method_frame.delivery_tag)
        return message

    def consume_message(self, processor):
        """
        Receive a message from the queue
        :param processor: A method to be invoked for each message, it receives
        a message and the channel in parameter and should return True to
        acknowledge the message/ False  to avoid doing so.
        """
        channel = self._channel

        def wrapper(channel, method_frame, header_frame, body):
            t = json.loads(body.decode('UTF-8'))
            message = Message(t, method_frame, header_frame)
            try:
                rc = processor(message)
                if rc:
                    channel.basic_ack(method_frame.delivery_tag)
            except StopIteration as e:
                # This is the normal exit method, we should consume the last
                # message so that it does not stay
                # the queue
                channel.basic_ack(method_frame.delivery_tag)
                channel.stop_consuming()

        channel.basic_consume(wrapper, queue=self._queue_name)
        try:
            channel.start_consuming()
        except InterruptedProcessing:
            channel.stop_consuming()
        except Exception as e:
            channel.stop_consuming()
            raise e


class AdminHelper(QueueHelper):
    """
       Class used for the administration helpers
    """

    def __init__(self, exchange, *args, **kwargs):
        """
        Initialize props
        :param exchange_name: the exchange name used in the broker
        """
        super(AdminHelper, self).__init__(*args, **kwargs)
        self._exchange_name = exchange

    def change_messages(self, queue_name, message_changer,
                        dest_exchange=None):
        """
        Method used to change the payload, the header (priority,
        max_retry_count, etc) or the exchange of message already in the queue

        :param queue_name: the source queue for the messages to change
        :param message_changer: the function called to change the message. It
                                should returned a tuple of (modified, message).
                                If the modified is False, the message is not
                                modified and will be not be ACK, thus from
                                the message broker point of view the message
                                will remain the same. If modified is true,
                                the original message will be consumed and the
                                new one will be re-enqueued.
        :param dest_exchange: if it is not None, the message will be send to
                              this exchange instead of the original one.
        """
        # First we declare the listening and sending brokers.
        # If we what to change the messages in the same queue, the sender
        # exchange is the same as the receiving one
        receive_broker = Receiver(self._exchange_name, queue_name,
                                  channel=self._channel, no_bind=True)
        if dest_exchange:
            sender_broker = Sender(dest_exchange, channel=self._channel)
        else:
            sender_broker = Sender(self._exchange_name, channel=self._channel)
        messages = []
        # We need to look at all the messages in the queue
        while True:
            # We receive each message one by one, but we do not send neither
            # ACK nor NACK ( noAck = true)
            message = receive_broker.receive_message(noAck=True)
            if message is None:
                break
            # We perform the message filtering and change
            (modified, message) = message_changer(message)

            # The messages are added in a list of message tuples that include
            # delivery tag for ACK or NACK, if it was modified (thus NACK is
            # needed) and the new message if modified
            if modified:

                routing_key = message.method_frame.routing_key
                max_retry = message.header_frame.headers.get('max_retry', 3)
                priority = message.header_frame.priority
                logging.info("[Message %s] %s (P:%s) was changed!" % (
                    message.method_frame.delivery_tag,
                    message.body, priority))

                messages.append({
                    'delivery_tag': message.method_frame.delivery_tag,
                    'has_been_modified': modified,
                    'updated_message': {
                        'routing_key': routing_key,
                        'payload': message.body,
                        'priority': priority,
                        'max_retry': max_retry
                    }})

            else:
                logging.info("[Message %s] %s (P:%s) was not changed!" % (
                    message.method_frame.delivery_tag,
                    message.body, message.header_frame.priority))
                messages.append({
                    'delivery_tag': message.method_frame.delivery_tag,
                    'has_been_modified': modified,
                    'updated_message': None
                })
        # If the message was modified, send first the new message into the
        # correct exchange and then ack the old message. If no updates were
        # done, just send a NACK for the existing message to leave it in the
        # queue.
        for message in messages:
            try:
                if message['has_been_modified']:
                    updated_message = message['updated_message']
                    sender_broker.send_message(updated_message['routing_key'],
                                               updated_message['payload'],
                                               priority=updated_message[
                                                   'priority'],
                                               max_retry=updated_message[
                                                   'max_retry'])
                    logging.info(
                        "[Message-%s] Sending ACK!" % message['delivery_tag'])
                    self._channel.basic_ack(message['delivery_tag'])
                else:
                    logging.info(
                        "[Message-%s] Sending NACK!" % message['delivery_tag'])
                    self._channel.basic_nack(message['delivery_tag'])
            except Exception as e:
                logging.error(
                    "Error processing message %s : %s. Sending NACK!" % (
                        message['delivery_tag'], str(e)))
                self._channel.basic_nack(message['delivery_tag'])

    def list_messages_in_queue(self, queue_name):
        """
        Display the messages in the queue
        :param queue_name: the queue whose message should be displayed
        :return: a list of tuples with the message details
        """
        receive_broker = Receiver(self._exchange_name, queue_name,
                                  channel=self._channel,
                                  no_bind=True)
        display = []
        messages = []
        try:
            while True:
                message = receive_broker.receive_message(noAck=True)
                if message is None:
                    break
                messages.append(message.method_frame.delivery_tag)
                routing_key = message.method_frame.routing_key
                max_retry = message.header_frame.headers.get('max_retry', None)
                priority = message.header_frame.priority
                try:
                    expire_in = message.header_frame.expiration
                except:
                    expire_in = None
                uid = message.header_frame.headers.get('uid', None)
                display.append([routing_key, max_retry, priority, expire_in,
                                uid] +
                               [str(x) for x in message.body])
        except Exception as e:
            logging.error("Error gathering messages info: %s" % e)
        finally:
            for message in messages:
                self._channel.basic_nack(message)
        return display


class RPCServer(Receiver):
    """
     RPC Server that is waiting for messages from a RPC Client on a given
     exchange and it replaying to the client anonymous queue.
     """

    def __init__(self, exchange, queue, routing_keys=None, auto_delete=False,
                 *args, **kwargs):
        """

        :param exchange_name: the exchange name used in the broker
        :param queue: the queue on which needs to be bind to the exchange
        :param routing_keys: the routing keys that are used to route a message
                             from the exchange to the queue
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        """
        super(RPCServer, self).__init__(exchange, queue, routing_keys,
                                        route_only_to_queue=True,
                                        auto_delete=auto_delete,
                                        *args, **kwargs)
        self.exchange = exchange
        self.auto_delete = auto_delete
        self.counter_messages = 0

    def receive_and_replay_rpc_message(self, processor, limit=None):
        """
        Method to consume a RPC message and to replay using the processor
        :param processor: the function used to process the consumed message
                          and generate the content of the replay message as
                          JSON message
        :param limit: the number of max message exchanges (receive and send)
        """
        def wrapper(message):
            self.counter_messages += 1
            replay_payload = json.dumps(processor(message))
            routing_key = message.header_frame.reply_to
            extra_properties = {
                'correlation_id': message.header_frame.correlation_id
            }
            self.basic_publish(self.exchange, routing_key, replay_payload,
                               priority=self.MAX_PRIORITY,
                               auto_delete=self.auto_delete,
                               extra_properties=extra_properties)
            if limit is not None and limit <= self.counter_messages:
                raise StopIteration
            return replay_payload is not None
        self.consume_message(wrapper)


class RPCClient(QueueHelper):
    """
    RPC Client that is sending messages to a RPC Server on a given exchange
    and it is waiting for the server replay on a anonymous queue.
    """

    def __init__(self, exchange, routing_keys=None, auto_delete=False,
                 *args, **kwargs):
        """
        :param exchange: the exchange name used in the broker
        :param routing_keys: the routing keys that are used to route a message
                             from the exchange to the queue
        :param auto_delete: option to automatically delete the exchange if no
                            queues is bind to the exchange
        """
        super(RPCClient, self).__init__(*args, **kwargs)
        self.auto_delete = auto_delete
        self.exchange_name = exchange
        self.corr_id = str(uuid.uuid4())
        self.queue_name = self.bind_queue_to_exchange(
            None, exchange_name=exchange,
            route_only_to_queue=True,
            routing_keys=routing_keys, auto_delete=auto_delete)
        self.counter_messages = 0

    def send_and_receive_rpc_message(self, routing_key, payload, processor,
                                     limit=None):
        """
        Method to send a RPC message and process the replay using the processor

        :param routing_keys: the routing keys that are used to route a message
                             from the exchange to the queue
        :param payload: the content of the message used in the RPC call
        :param processor: the function used to process the consumed message
        :param limit: the number of max message exchanges (send and receive)
        """
        body = json.dumps(payload)
        channel = self._channel
        extra_properties = {
            'correlation_id': self.corr_id,
            'reply_to': self.queue_name
        }

        def wrapper(channel, method_frame, header_frame, body):
            self.counter_messages += 1
            t = json.loads(body.decode('UTF-8'))
            message = Message(t, method_frame, header_frame)
            if self.corr_id != header_frame.correlation_id:
                channel.basic_ack(method_frame.delivery_tag)
            else:
                rc = processor(message)
                if rc:
                    channel.basic_ack(method_frame.delivery_tag)
                else:
                    channel.basic_nack(method_frame.delivery_tag)
            if limit is not None and limit <= self.counter_messages:
                channel.stop_consuming()

        channel.basic_consume(wrapper, queue=self.queue_name)
        self.basic_publish(self.exchange_name, routing_key, body,
                           priority=self.MAX_PRIORITY,
                           auto_delete=self.auto_delete,
                           extra_properties=extra_properties
                           )
        try:
            channel.start_consuming()
        except InterruptedProcessing:
            channel.stop_consuming()
        except Exception as e:
            channel.stop_consuming()
            raise e



from lbmessaging.services.Common import BaseService


class CVMFSNightliesService(BaseService):
    """
    CVMFS Nightlies installation service
    The exposed exchanges and their methods are:
        - dev_send              mapped to send_command of CvmfsDevExchange
        - dev_receive           mapped to receive_command of CvmfsDevExchange
        - conddb_send           mapped to send_command of CvmfsConDBExchange
        - conddb_receive        mapped to receive_command of CvmfsConDBExchange
        - prod_send             mapped to send_command of CvmfsProdExchange
        - prod_receive          mapped to receive_command of CvmfsProdExchange
        - send_build_ready      mapped to send_build_ready of
                                ContinuousIntegrationExchange
        - receive_build_ready   mapped to receive_build_ready of
                                ContinuousIntegrationExchange
    """
    def __init__(self, host=None, username=None, passwd=None, port=5671,
                 vhost='/lhcb', use_ssl=True):
        """
        :param host: the RabbitMQ server hostname
        :param username: the RabbitMQ server username
        :param passwd: the RabbitMQ server password
        :param port: the RabbitMQ server port
        :param vhost: the RabbitMQ virtual host that should be used in the
                      connection
        :param use_ssl: if True, an encrypted connection is created, plain text
                        otherwise
        """
        super(CVMFSNightliesService, self).__init__(host=host,
                                                    username=username,
                                                    passwd=passwd, port=port,
                                                    vhost=vhost,
                                                    use_ssl=use_ssl)



###############################################################################
# (c) Copyright 2017 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Module in charge of dealing with sending and receiving perioding tests messages
 for the PR2
'''

import json
from collections import namedtuple

from lbmessaging.exchanges.Common import Sender, Receiver, Message, \
    check_channel

__author__ = 'Chitic Stefan-Gabriel <stefan-gabriel.chitic@cern.ch>'

PERIODIC_TESTS_EXCHANGE = "topic.periodic_test"

# Utility class to ease the use of the messages
PeriodicTestMessage = namedtuple('PeriodicTestMessage',
                                 ['slot', 'build_id', 'project', 'platform',
                                  'group', 'env', 'runner', 'os_label'])


def _message_to_periodic(message):
    """ Trivial utility that converts the periodic test from JSON
    :param message: the message that needs to be converted to command
    :returns the converted rabbitMQ message to Message object with payload as
             command message
    """
    ret = None
    if message and message.body:
        try:
            msg = json.loads(message.body)
        except:
            msg = message.body
        if isinstance(msg, list):
            msg = msg[0]
        try:
            body = PeriodicTestMessage(**msg)
        except TypeError:
            # Add missing fields in the body to None
            for field in PeriodicTestMessage._fields:
                if not msg.get(field, None):
                    msg[field] = None
            body = PeriodicTestMessage(**msg)

        ret = Message(body, message.method_frame, message.header_frame)
    return ret


def command_retry_count(message):
    """ returns the retry_count kept in the headers
    :param message: the message whose retries needs to be counted
    :returns the value of the retry_count from the header
    """
    return message.header_frame.headers.get('max_retry', None)


class PeriodicTestsExchange(object):
    """
    Class in charge of dealing with messages to and from the PR insance
    """

    def __init__(self, channel, connection=None):
        """
        Specify the exchange to be used
        :param channel: a pikaBlockingConnection channel that should be used in
                        the broker
        :param connection: a pikaBlockingConnection that should be used in
                        the broker
        """
        self._exchange = PERIODIC_TESTS_EXCHANGE

        self._channel = check_channel(channel, connection)

    def request_test(self, slot, buildId, project, config, group, env,
                     runner, os_label, priority=10, max_retry=3,
                     expiration=None):
        """ Formats and sends the command as JSON
        :param command: the command that should be executed on a CVMFS instance
        :param arguments: the command arguments
        :param priority: the priority of the message
        :param max_retry: the max of retries in case of command failure on CVMFS
        :param expiration: the TTL of the message
        """
        params = [slot, project, config, group, env]
        routingKey = ".".join(params)
        body = {
            'slot': slot, 'build_id': buildId, 'project': project,
            'platform': config, 'group': group, 'env': env,
            'runner': runner, 'os_label': os_label}
        s = Sender(self._exchange, channel=self._channel)
        s.send_message(routingKey, body, priority,
                       expiration=expiration,
                       max_retry=max_retry)

    def get_test_to_run(self, queue_name, callback=None):
        """ Receive one message from the queue
        :param queue_name: the queue on which the broker listens for messages
        :param callback: function to run before sending the ack
        :returns the last message converted to Message object with payload as
                 command message
        """
        r = Receiver(self._exchange, queue_name,
                     channel=self._channel)
        message = r.receive_message(runBeforeAck=callback)
        return _message_to_periodic(message)

    def get_tests_to_run(self, queue_name, callback=None):
        """ Receive all message from the queue with callback
        :param queue_name: the queue on which the broker listens for messages
        :param callback: function to run before sending the ack
        :returns the last message converted to Message object with payload as
                 command message
        """
        test_list = []
        while True:
            message = self.get_test_to_run(queue_name, callback=callback)
            if not message:
                break
            test_list.append(message)
        return test_list



import logging
import multiprocessing
import signal
import sys
import unittest
from time import sleep

import os
from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.Common import Sender, Receiver, InterruptedProcessing

EXCHANGE = "test_exchange_signal"
QUEUE = "test_queue_signal"


def run_consumer(time_delay):
    logging.warning("Starting consume")
    connection = get_test_connection()

    r = Receiver(EXCHANGE, QUEUE, connection=connection)

    # Added signal handler
    def signal_handler(sig, frame):
        logging.warning("Received signal %s" % s)
        if sig in [ signal.SIGTERM,  signal.SIGINT ]:
            raise InterruptedProcessing("Controlled stop: %s" % sig)
        else:
            raise Exception("Uncontrolled stop: %s" % sig)

    for s in [ signal.SIGTERM, signal.SIGINT ]:
        signal.signal(s, signal_handler)

    # Now receiving...
    receivedlist = []
    def myconsumer(message):
        logging.warning("Processing:" + str(message.body))
        # Make sure we still sleep when the other process interrupts us...
        sleep(time_delay)
        receivedlist.append(message.body)
        return True

    try:
        msgs = r.consume_message(myconsumer)
    except:
        logging.warning("After consuming message")
        logging.debug(receivedlist)

class TestSignal(unittest.TestCase):


    def setUp(self):
        self._connection = get_test_connection()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_signal_stop(self):

        # Consume messages left over from previous runs
        r = Receiver(EXCHANGE, QUEUE, connection=self._connection)
        r.receive_messages()

        # Now enqueuing
        logging.warning("Sending message to test server stop")
        txt = "LONG_RUNNING_TEST"
        self.send(txt)

        # Starting a child that will consume the messages
        time_delay = 10000
        logging.warning("Starting consumer process")
        p = multiprocessing.Process(target=run_consumer, args=([ time_delay ]))
        p.start()
        sleep(1)

        # sending SIGTERM to child
        os.kill(p.pid, signal.SIGTERM)

        # Waiting for child to finish
        p.join()
        self.assertEqual(0, p.exitcode)

        # The message should still be in the queue !
        m = r.receive_message()
        self.assertIsNotNone(m)
        self.assertEqual(txt, m.body)

    def test_signal_crash(self):

        # Consume messages left over from previous runs
        r = Receiver(EXCHANGE, QUEUE, connection=self._connection)
        r.receive_messages()

        # Now enqueuing
        logging.warning("Sending message to test crash")
        txt = "LONG_RUNNING_TEST"
        self.send(txt)

        # Starting a child that will consume the messages
        time_delay = 10000
        logging.warning("Starting consumer process")
        p = multiprocessing.Process(target=run_consumer, args=([ time_delay ]))
        p.start()
        sleep(1)

        # sending SIGABRT to child
        os.kill(p.pid, signal.SIGABRT)

        # Waiting for child to finish
        p.join()
        self.assertNotEquals(0, p.exitcode)

        # The message should still be in the queue !
        m = r.receive_message()
        self.assertIsNotNone(m)
        self.assertEqual(txt, m.body)


    def send(self, message_text):
        s = Sender(EXCHANGE, connection=self._connection)
        s.send_message("test", message_text)



if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



from lbmessaging.services.Common import BaseService
import os

class LbCIService(BaseService):
    """
    LHCb continuous integration messaging service
    The exposed exchanges and their methods are:

        - send_build_ready      mapped to send_build_ready of
                                ContinuousIntegrationExchange
        - consume_build_ready   mapped to consume_build_ready of
                                ContinuousIntegrationExchange
        - receive_build_ready   mapped to receive_build_ready of
                                ContinuousIntegrationExchange
        - send_envkit_ready     mapped to send_envkit_ready of
                                ContinuousIntegrationExchange
        - consume_envkit_ready  mapped to consume_envkit_ready of
                                ContinuousIntegrationExchange
        - receive_envkit_ready  mapped to receive_envkit_ready of
                                ContinuousIntegrationExchange
    """
    def __init__(self, host=None, username=None, passwd=None, port=5671,
                 vhost='/lhcb', use_ssl=True):
        """
        :param host: the RabbitMQ server hostname
        :param username: the RabbitMQ server username
        :param passwd: the RabbitMQ server password
        :param port: the RabbitMQ server port
        :param vhost: the RabbitMQ virtual host that should be used in the
                      connection
        :param use_ssl: if True, an encrypted connection is created, plain text
                        otherwise
        """
        super(LbCIService, self).__init__(host=host, username=username,
                                          passwd=passwd, port=port,
                                          vhost=vhost, use_ssl=use_ssl)

    @staticmethod
    def getEnvKitUrl(flavour, platform, version, url=None,
                     reponame='lhcbdev.cern.ch'):
        if url:
            return url
        base = {
            'lhcb.cern.ch': 'lbenv-prod-kit',
        }.get(reponame, 'lbenv-kit')
        return "http://lhcb-rpm.web.cern.ch/lhcb-rpm/%ss/" \
               "%s.%s.%s.%s.tar.bz2" % (base, base, flavour, platform, version)

    @staticmethod
    def getEnvKitPrefix(flavour, platform, reponame='lhcbdev.cern.ch'):
        env_kit_top_dir = {
            'lhcb.cern.ch': '/cvmfs/lhcb.cern.ch/lib/var/lib/LbEnv',
        }.get(reponame, '/cvmfs/lhcbdev.cern.ch/new-msr/var/lib/LbEnv')
        path = os.path.normpath(
            os.path.join(env_kit_top_dir, flavour, platform))
        if not path.startswith(env_kit_top_dir):
            raise ValueError("Trying to compute a path outside the top "
                             "directory for envkit: %s" % path)
        return path + os.path.sep



import logging
import sys
import unittest

from lbmessaging.exchanges.CvmfsDevExchange import CvmfsDevExchange, \
    CVMFSDEV_EXCHANGE
from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.Common import AdminHelper, Receiver

QUEUE_NAME = "test_cvmfs_actions_queue"
QUEUE_NAME_CHANGED = "test_cvmfs_actions_queue_changed"
NEW_EXCHANGE = "cvmfsdev.action_backup"


class TestCvmfsDevMsg(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._channel = self._connection.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def _update_priority(self, message):
        if 'test_cmd_5' in message[0]:
            message[2].priority = 30
            return (True, message)
        return (False, message)

    def test_display(self):
        """ Check whether a the admin helpers display correctly the messages
        """
        broker = CvmfsDevExchange(self._channel)
        broker.receive_all(QUEUE_NAME)

        command = "test_cmd_%s"
        args = ["arg1", "arg2"]
        expected_results = []
        for i in range(0, 10):
            broker.send_command(command % i, args)
            expected_results.append([
                command % i, 3, 10, None, command % i,
                "['arg1', 'arg2']"
            ])
        # Test simple display
        tools = AdminHelper(CVMFSDEV_EXCHANGE,
                            channel=self._channel)
        res = tools.list_messages_in_queue(QUEUE_NAME)
        results_without_time = []
        for r in res:
            results_without_time.append([
                r[0], r[1], r[2], r[3], r[5], r[6].replace('u\'', '\'')
            ])
            self.assertTrue(r[4] is not None)

        self.assertEqual(expected_results, results_without_time)

        # Verify if all the messages are still in queue
        res = broker.receive_all(QUEUE_NAME)
        self.assertEqual(len(res), 10)

    def test_message_change(self):
        """ Check whether a the admin helpers display correctly the messages
        """
        broker = CvmfsDevExchange(self._channel)
        broker.receive_all(QUEUE_NAME)

        command = "test_cmd_%s"
        args = ["arg1", "arg2"]
        expected_results = []
        expected_results.append([
            command % 5, 3, 30, None, command % 5,
            "['arg1', 'arg2']"
        ])
        counter = 9
        for i in range(0, 10):
            broker.send_command(command % i, args)
            if i == 5:
                continue
            counter -= 1
            expected_results.append([
                command % i, 3, 10, None, command % i,
                "['arg1', 'arg2']"
            ])
        # Test simple display
        tools = AdminHelper(CVMFSDEV_EXCHANGE,
                            channel=self._channel)
        tools.change_messages(QUEUE_NAME, self._update_priority)
        res = tools.list_messages_in_queue(QUEUE_NAME)
        results_without_time = []
        for r in res:
            results_without_time.append([
                r[0], r[1], r[2], r[3], r[5], r[6].replace('u\'', '\'')
            ])
            self.assertTrue(r[4] is not None)

        self.assertEqual(expected_results, results_without_time)

        # Verify if all the messages are still in queue
        res = broker.receive_all(QUEUE_NAME)
        self.assertEqual(len(res), 10)

    def test_exchange_change(self):
        """ Check whether a the admin helpers display correctly the messages
        """
        broker = CvmfsDevExchange(self._channel)
        broker.receive_all(QUEUE_NAME)

        new_broker = Receiver(NEW_EXCHANGE,
                              QUEUE_NAME_CHANGED, channel=self._channel)
        # Consume messages left over from previous runs
        new_broker.receive_messages()

        command = "test_cmd_%s"
        args = ["arg1", "arg2"]
        expected_results = []
        counter = 9
        for i in range(0, 10):
            broker.send_command(command % i, args)
            if i == 5:
                continue
            counter -= 1
            expected_results.append([
                command % i, 3, 10, None, command % i,
                "['arg1', 'arg2']"
            ])
        # Test simple display
        tools = AdminHelper(CVMFSDEV_EXCHANGE,
                            channel=self._channel)
        tools.change_messages(QUEUE_NAME, self._update_priority,
                              dest_exchange=NEW_EXCHANGE)
        res = tools.list_messages_in_queue(QUEUE_NAME)
        results_without_time = []
        for r in res:
            results_without_time.append([
                r[0], r[1], r[2], r[3], r[5], r[6].replace('u\'', '\'')
            ])
            self.assertTrue(r[4] is not None)

        self.assertEqual(expected_results, results_without_time)

        # Verify if the 5th message is in the other exchange
        msgs = new_broker.receive_messages()
        self.assertEqual(len(msgs), 1)
        self.assertEqual(str(msgs[0][0][0]), "test_cmd_5")
        # Verify if all the messages are still in queue
        res = broker.receive_all(QUEUE_NAME)
        self.assertEqual(len(res), 9)

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



import logging
import sys
import threading
import time
import unittest

from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.Common import RPCServer, RPCClient

QUEUE_NAME = "test_rpc"
EXCHANGE_NAME = 'cvmfs.test.rpc'


class TestRPC(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._connection_server = get_test_connection()
        self._channel = self._connection.channel()
        self._channel_server = self._connection_server.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def _trigger_server(self):
        server = RPCServer(EXCHANGE_NAME, QUEUE_NAME,
                           channel=self._channel_server,
                           auto_delete=True)

        def wrapper(message):
            self.assertEqual(message.body['question'], 'Done?')
            return {'result': 'Done'}

        server.receive_and_replay_rpc_message(wrapper, limit=1)

    def test_rpc(self):
        """ Check whether a simple command gets through """
        server_thread = threading.Thread(target=self._trigger_server)
        server_thread.start()
        time.sleep(1)
        client = RPCClient(EXCHANGE_NAME,
                           routing_keys=None,
                           auto_delete=True,
                           channel=self._channel)

        def wrapper(message):
            self.assertEqual(message.body['result'], 'Done')
            return True

        client.send_and_receive_rpc_message(QUEUE_NAME,
                                            {'question': 'Done?'},
                                            wrapper, limit=1)
        server_thread.join()

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



import logging
import sys
import time
import unittest

from lbmessaging.exchanges.CvmfsDevExchange import CvmfsDevExchange
from lbmessaging.exchanges.Common import get_connection

from pika import exceptions

QUEUE_NAME = "test_cvmfs_actions_queue"
ERROR_QUEUE_NAME = "test_cvmfs_actions_queue_errors"

class TestTimeout(unittest.TestCase):

    def get_broker(self):
        self._connection = get_connection(vhost='/lhcb-test',
                                          heartbeat_interval=1,
                                          blocked_connection_timeout=5)
        self._channel = self._connection.channel()
        broker = CvmfsDevExchange(self._channel)
        return broker

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_connection_closed(self):
        """ Check what happens when the server closes the connection """
        command = "test_cmd"
        args = ["arg1", "arg2"]
        # purge the queue
        broker = self.get_broker()
        broker.receive_all(QUEUE_NAME)
        logging.info("Sending 2 messages")
        broker.send_command(command, args)
        broker.send_command(command, args)
        logging.info("receiving the 1st")
        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(command, tmp.body.command)
        self.assertEqual(args, tmp.body.arguments)
        logging.info("Now waiting 10 seconds,"
                     " which should cause the conenction to timeout")
        time.sleep(10)
        try:
            logging.info("receiving the 2nd")
            tmp = broker.receive_command(QUEUE_NAME)
        except exceptions.ConnectionClosed as cc:
            logging.info("Connection was closed, trying with a new one...")
            new_broker = self.get_broker()
            tmp = new_broker.receive_command(QUEUE_NAME)
            self.assertEqual(command, tmp.body.command)
            self.assertEqual(args, tmp.body.arguments)

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



import logging
import sys
import unittest

from lbmessaging.services.CVMFSNightliesService import CVMFSNightliesService
from lbmessaging.exchanges.CvmfsDevExchange import CvmfsDevExchange
from lbmessaging.exchanges.test import get_test_connection


QUEUE_NAME = "test_cvmfs_actions_queue"


class TestCVMFSNightliesService(unittest.TestCase):

    def setUp(self):
        channel = get_test_connection().channel()
        broker = CvmfsDevExchange(channel)
        broker.receive_all(QUEUE_NAME)

    def tearDown(self):
        pass

    def test_services(self):
        command = "test_cmd"
        args = ["arg1", "arg2"]
        srv = CVMFSNightliesService(vhost='/lhcb-test')
        srv.dev_send(command, args)
        tmp = srv.dev_receive(QUEUE_NAME)
        self.assertEqual(command, tmp.body.command)
        self.assertEqual(args, tmp.body.arguments)

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



###############################################################################
# (c) Copyright 2017 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Module in charge of dealing with sending and receiving nightlies information to
RabbitMQ when a project is done: slot, build id, platform and project
'''

import json
from collections import namedtuple

from lbmessaging.exchanges.Common import Sender, Receiver, Message,\
    check_channel
from lbmessaging.services.Common import ExposedIn


__author__ = 'Stefan-Gabriel Chitic <stefan-gabriel.chitic@cern.ch>'


PHYS_CI_EXCHANGE = "topic.build_ready"
ENVKIT_CI_EXCHANGE = "topic.tar_ready"


# Utility class to ease the use of the messages
PhysBuildReadyMessage = namedtuple('PhysBuildReadyMessage',
                                   sorted(['slot', 'build_id',
                                           'platform', 'project',
                                           'priority',
                                           'deployment']),
                                   rename=True)
EnvkitReadyMessage = namedtuple('EnvkitReadyMessage',
                                sorted(['flavour', 'platform', 'version',
                                        'url']),
                                rename=True)


def _message_to_exchange_type_message(message, type=PhysBuildReadyMessage):
    """ Trivial utility that converts the command from JSON
    :param message: the message that needs to be converted to command
    :param type: the type of message to which the payload is converted.
                 Default is PhysBuildReadyMessage
    :returns the converted rabbitMQ message to Message object with payload as
             ContinuousIntegrationMessage """
    ret = None
    if message and message.body:
        try:
            msg = json.loads(message.body)
        except:
            msg = message.body
        if isinstance(msg, list):
            msg = msg[0]
        try:
            body = type(**msg)
        except TypeError:
            # Add missing fields in the body to None
            for field in type._fields:
                if not msg.get(field, None):
                    msg[field] = None
            body = type(**msg)

        ret = Message(body, message.method_frame, message.header_frame)
    return ret


def message_retry_count(message):
    """ returns the retry_count kepts in the headers
    :param message: the message whose retries needs to be counted
    :returns the value of the retry_count from the header
    """
    return message.header_frame.headers.get('max_retry', None)


@ExposedIn('LbCIService', mapping={
        'send_build_ready': 'send_build_ready',
        'consume_build_ready': 'consume_build_ready',
        'receive_build_ready': 'receive_build_ready',
        'send_envkit_ready': 'send_envkit_ready',
        'consume_envkit_ready': 'consume_envkit_ready',
        'receive_envkit_ready': 'receive_envkit_ready',
    })
@ExposedIn('CVMFSNightliesService', mapping={
        'send_build_ready': 'send_build_ready',
        'receive_build_ready': 'receive_build_ready',
    })
class ContinuousIntegrationExchange(object):
    """
    Class in charge of dealing with messages to and from the
    ContinuousIntegration Exchange
    """

    def __init__(self, channel, connection=None):
        """
        :param channel: a pikaBlockingConnection channel that should be used in
                        the broker
        :param connection: a pikaBlockingConnection that should be used in
                        the broker"""
        self._channel = check_channel(channel, connection)

    def send_envkit_ready(self, flavour, platform, version, url=None):
        """ Formats and sends the information for a tar that was build as JSON
        :param flavour: The flavour of the env_kit (prod/dev)
        :param platform: The platform of the env_kit
        :param version: the version of the env_kit
        :param url: Optional url to override the source of the env_kit
        :return:
        :param priority: the priority of the message
        """
        message = {
            'flavour': flavour,
            'platform': platform,
            'version': version,
            'url': url
        }
        routing_key = "%s.%s.%s.%s" % (flavour, platform, version, url)
        self._send_continuous_integration(ENVKIT_CI_EXCHANGE, message,
                                          routing_key)

    def receive_envkit_ready(self, queue_name):
        """ Receive one tar ready message from the queue
        :param queue_name: the queue on which the broker listens for messages
        :returns the last message converted to Message object with payload as
                 envkitReadyMessage
        """
        return self._receive_continuous_integration(
            queue_name, ENVKIT_CI_EXCHANGE, message_type=
            EnvkitReadyMessage)

    def receive_all_envkit_ready(self, queue_name):
        """ Receive all tar ready messages from the queue
        :param queue_name: the queue on which the broker listens for messages
        :returns a list of all the messages converted to Message object with
                 payload as EnvkitReadyMessage
        """
        return self._receive_continuous_integration_all(
            queue_name, ENVKIT_CI_EXCHANGE,
            message_type=EnvkitReadyMessage)

    def consume_envkit_ready(self, message_processor, queue_name):
        """ Invokes the message_processor on all messages received on
        envkit_CI_EXCHANGE.
         The message processor should:
         - take (message, channel) as arguments
         - return True to ack the messages

         It can interrupt the loop by calling channel.stop_consuming()

         BEWARE: This is not appropriate for long running calls as the
         acknowledgement of the message is done after the processor call.

        :param message_processor: the function that should be called when a new
                                  message is consumed
        :param queue_name: the queue on which the broker listens for messages
        """
        self._consume_continuous_integration(message_processor, queue_name,
                                             ENVKIT_CI_EXCHANGE,
                                             message_type=EnvkitReadyMessage)

    def send_build_ready(self, slot, build_id, platform, project,
                         deployment=[], priority=None):
        """ Formats and sends the a pair of slot-build_id-platform-project
         that was build as JSON
        :param slot: the slot name that was build
        :param build_id: the build id of the slot that was build
        :param platform: the platform of the slot that was build
        :param project: the project of the slot that was build
        :param priority: the priority of the message
        """
        message = {
            'slot': slot,
            'build_id': build_id,
            'project': project,
            'platform': platform,
            'deployment': deployment,
            'priority': priority
        }
        routing_key = "%s.%s.%s.%s" % (slot, build_id, project, platform)
        self._send_continuous_integration(PHYS_CI_EXCHANGE, message,
                                          routing_key)

    def receive_build_ready(self, queue_name):
        """ Receive one build ready message from the queue
        :param queue_name: the queue on which the broker listens for messages
        :returns the last message converted to Message object with payload as
                 PhysBuildReadyMessage
        """
        return self._receive_continuous_integration(queue_name,
                                                    PHYS_CI_EXCHANGE)

    def receive_all_build_ready(self, queue_name):
        """ Receive all build ready messages from the queue
        :param queue_name: the queue on which the broker listens for messages
        :returns a list of all the messages converted to Message object with
                 payload as PhysBuildReadyMessage
        """
        return self._receive_continuous_integration_all(queue_name,
                                                        PHYS_CI_EXCHANGE)

    def consume_build_ready(self, message_processor, queue_name):
        """ Invokes the message_processor on all messages received on
        envkit_CI_EXCHANGE.
         The message processor should:
         - take (message, channel) as arguments
         - return True to ack the messages

         It can interrupt the loop by calling channel.stop_consuming()

         BEWARE: This is not appropriate for long running calls as the
         acknowledgement of the message is done after the processor call.

        :param message_processor: the function that should be called when a new
                                  message is consumed
        :param queue_name: the queue on which the broker listens for messages
        """
        self._consume_continuous_integration(message_processor, queue_name,
                                             PHYS_CI_EXCHANGE)

    def _send_continuous_integration(self, exchange, message, routing_key):
        """ Sends the a continuous integration object metainfo that was build
        as JSON
        :param exchange: the exchange that should be used
        :param message: the actual message that is sent
        :param routing_key: the routing key of the message
         """
        s = Sender(exchange,
                   channel=self._channel)
        s.send_message(routing_key, message, priority=0,
                       max_retry=0)

    def _receive_continuous_integration(self, queue_name, exchange,
                                        message_type=PhysBuildReadyMessage):
        """ Receive one message from the queue
        :param queue_name: the queue on which the broker listens for messages
        :param exchange: the exchange that should be used to map the queue to
        :param type: the type of message to which the payload is converted.
                 Default is PhysBuildReadyMessage
        :returns the last message converted to Message object with payload as
                 message_type
        """
        r = Receiver(exchange, queue_name,
                     channel=self._channel)
        message = r.receive_message(runBeforeAck=None)
        return _message_to_exchange_type_message(message, type=message_type)

    def _receive_continuous_integration_all(self, queue_name, exchange,
                                            message_type=PhysBuildReadyMessage):
        """ Receive all messages from the queue
        :param queue_name: the queue on which the broker listens for messages
        :param exchange: the exchange that should be used to map the queue to
        :param type: the type of message to which the payload is converted.
                 Default is PhysBuildReadyMessage
        :returns a list of all the messages converted to Message object with
                 payload as message_type
        """
        r = Receiver(exchange, queue_name,
                     channel=self._channel)
        msglist = r.receive_messages()
        return [_message_to_exchange_type_message(m, type=message_type)
                for m in msglist]

    def _consume_continuous_integration(self, message_processor, queue_name,
                                        exchange,
                                        message_type=PhysBuildReadyMessage):
        """ Invokes the message_processor on all messages received.
         The message processor should:
         - take (message, channel) as arguments
         - return True to ack the messages

         It can interrupt the loop by calling channel.stop_consuming()

         BEWARE: This is not appropriate for long running calls as the
         acknowledgement of the message is done after the processor call.

        :param message_processor: the function that should be called when a new
                                  message is consumed
        :param queue_name: the queue on which the broker listens for messages
        :param exchange: the exchange that should be used to map the queue to
        :param type: the type of message to which the payload is converted.
                 Default is PhysBuildReadyMessage
         """
        c = Receiver(exchange, queue_name,
                     channel=self._channel)
        # Adding wrapper to convert the messages before processing

        def convert_and_process(message):
            return message_processor(
                _message_to_exchange_type_message(message, type=message_type))
        c.consume_message(convert_and_process)



import logging
import sys
import unittest

from lbmessaging.exchanges.PeriodicTestsExchange import PeriodicTestsExchange
from lbmessaging.exchanges.test import get_test_connection


QUEUE_NAME = "test_pr_test_queue"


class TestCvmfsDevMsg(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._channel = self._connection.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_simple(self):
        """ Check whether a simple command gets through """
        slot = 'slot1'
        build_id = 'buildId1'
        project = 'project1'
        platform = 'platform1'
        group = 'group1'
        env = 'env1'
        runner = 'runner1'
        os_label = 'os_label1'
        broker = PeriodicTestsExchange(self._channel)
        # purge the queue
        broker.get_tests_to_run(QUEUE_NAME)

        broker.request_test(slot, build_id, project,
                            platform, group, env, runner, os_label)
        tmp = broker.get_test_to_run(QUEUE_NAME)
        self.assertEqual(slot, tmp.body.slot)
        self.assertEqual(build_id, tmp.body.build_id)
        self.assertEqual(project, tmp.body.project)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(group, tmp.body.group)
        self.assertEqual(env, tmp.body.env)
        self.assertEqual(runner, tmp.body.runner)
        self.assertEqual(os_label, tmp.body.os_label)

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



import json
import logging
import random
import sys
import unittest

from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.Common import Sender, Receiver


class TestConsumer(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._channel = self._connection.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_consumer(self):

        exchange = "test_exchange"
        queue = "test_queue"

        s = Sender(exchange, channel=self._channel)
        r = Receiver(exchange, queue, channel=self._channel)

        # Consume messages left over from previous runs
        r.receive_messages()

        mylist = list(range(5))
        random.shuffle(mylist)

        # Sending some messages
        for i in mylist:
            s.send_message("test", json.dumps([i]), priority=i+1)
        # Plus one to tell the process to stop consuming messages...
        s.send_message("test", "EXIT")

        # Now receiving...
        receivedlist = []
        def myconsumer(message):
            if message.body == "EXIT":
                raise StopIteration
            receivedlist.append(message.body)
            return True

        msgs = r.consume_message(myconsumer)

        logging.debug(receivedlist)
        # Checking that we have the same number of messages and that
        # they are sorted correctly...
        self.assertEqual(len(mylist), len(receivedlist))
        self.assertEqual(sorted(receivedlist)[::-1], receivedlist)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



###############################################################################
# (c) Copyright 2017 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Package initialization
"""

from lbmessaging.exchanges.Common import get_connection

# Priority management
##################################

# Levels
EMERGENCY = "EMERGENCY"
HIGH = "HIGH"
NORMAL = "NORMAL"
LOW = "LOW"

# Their mappings to actual values
# We add an extra level for the sake of the priority computation
_levels = [LOW, NORMAL, HIGH, EMERGENCY, "END"]
_values = [0, 128, 200, 255, 256]
_map = dict(zip(_levels, _values))


def priority(level, weight=0.0):
    """
    :param level: The priority level for the message, should be one of the official levels
    :param weight: important withn the range (float within 0 and 1)
    :return: the priority for the message translated to the range known by RabbitMQ queues (0-255)
    """

    if (level not in _levels):
        raise Exception("Unknown level")

    # +1 works as we have an extra / undeclared level
    nextlevel = _levels[_levels.index(level) + 1]
    low_priority = _map[level]
    high_priority = _map[nextlevel] - 1

    # map linearly to the range...
    return int(low_priority + weight * float(high_priority - low_priority))



import logging
import sys
import unittest

from lbmessaging.exchanges.CommandExchange import command_retry_count
from lbmessaging.exchanges.CvmfsDevExchange import CvmfsDevExchange
from lbmessaging.exchanges.CvmfsProdExchange import CvmfsProdExchange
from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.CvmfsConDBExchange import CvmfsConDBExchange

QUEUE_NAME = "test_cvmfs_actions_queue"


class TestCvmfsDevMsg(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._channel = self._connection.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_simple(self):
        """ Check whether a simple command gets through """
        command = "test_cmd"
        args = ["arg1", "arg2"]
        broker = CvmfsDevExchange(self._channel)
        # purge the queue
        broker.receive_all(QUEUE_NAME)

        broker.send_command(command, args)
        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(command, tmp.body.command)
        self.assertEqual(args, tmp.body.arguments)

    def test_condb(self):
        """ Check whether the cvmfs cond exchange works """
        command = "test_cmd"
        args = ["arg1", "arg2"]
        broker = CvmfsConDBExchange(self._channel)
        # purge queue first
        broker.receive_all("test_cvmfscondb_actions_queue")
        broker.send_command(command, args)
        tmp = broker.receive_command("test_cvmfscondb_actions_queue")
        self.assertEqual(command, tmp.body.command)
        self.assertEqual(args, tmp.body.arguments)

    def test_prod(self):
        """ Check the cvmfs prod exchange works """
        command = "test_cmd"
        args = ["arg1", "arg2"]
        broker = CvmfsProdExchange(self._channel)
        # purge queue first
        broker.receive_all("test_cvmfsprod_actions_queue")
        broker.send_command(command, args)
        tmp = broker.receive_command("test_cvmfsprod_actions_queue")

        self.assertEqual(command, tmp.body.command)
        self.assertEqual(args, tmp.body.arguments)

    def test_priority(self):
        """ Check the priority system """
        broker = CvmfsDevExchange(self._channel)

        # purge the queue
        broker.receive_all(QUEUE_NAME)

        # Now adding our commands
        cmds_low = ["low1", "low2", "low3"]
        cmd_high = "high1"
        for cmd in cmds_low:
            broker.send_command(cmd, [], 2)

        # Make sure we get the first command
        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(cmds_low[0], tmp.body.command)

        # Now enqueue a high priority command and check it's coming
        # before the others
        broker.send_command(cmd_high, [], 100)
        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(cmd_high, tmp.body.command)

        # Now check that we have the others too
        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(cmds_low[1], tmp.body.command)

        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(cmds_low[2], tmp.body.command)

    def test_error(self):
        """ Check the priority system """
        broker = CvmfsDevExchange(self._channel)

        # purge the queue
        broker.receive_all(QUEUE_NAME)
        broker.receive_command_error_all(QUEUE_NAME)

        # Now adding our command
        cmd = "test_command"
        broker.send_command(cmd, [], 100, max_retry=1)
        tmp = broker.receive_command(QUEUE_NAME)
        self.assertEqual(cmd, tmp.body.command)

        # Pretend we have a processing error
        broker.handle_processing_error(tmp, queue_name=QUEUE_NAME)
        tmp2 = broker.receive_command(QUEUE_NAME)
        self.assertEqual(0, command_retry_count(tmp2))

        # Pretend we have a processing error again
        broker.handle_processing_error(tmp2, queue_name=QUEUE_NAME)
        tmp3 = broker.receive_command_error(QUEUE_NAME)
        self.assertEqual(cmd, tmp3.body.command)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



import importlib
import pkgutil

"""
Imports all the submodules in order to create their exposure in the annotated
services
"""
for _, name, __ in pkgutil.walk_packages(__path__,
                                         prefix=__name__ + '.'):
    if 'test' in name:
        continue
    importlib.import_module(name)



import json
import logging
import sys
import time
import unittest

from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.Common import Sender, Receiver


class TestPriorityQueue(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._channel = self._connection.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_existing(self):

        exchange = "test_ttl"
        queue = "test_queue_ttl"

        s = Sender(exchange, channel=self._channel)
        r = Receiver(exchange, queue, channel=self._channel)

        # Consume messages left over from previous runs
        r.receive_messages()

        mylist = list(range(5))

        for i in mylist:
            s.send_message("test", json.dumps([i]), expiration=1000)

        receivedlist = []
        msg = r.receive_message()
        logging.warning(msg)
        while msg:
            o = json.loads(msg[0])
            receivedlist.append(o[0])
            msg = r.receive_message()

        logging.warning(receivedlist)
        self.assertEqual(len(mylist), len(receivedlist))

    def test_expired(self):

        exchange = "test_ttl"
        queue = "test_queue_ttl"

        s = Sender(exchange, channel=self._channel)
        r = Receiver(exchange, queue, channel=self._channel)

        # Consume messages left over from previous runs
        r.receive_messages()

        mylist = list(range(5))

        for i in mylist:
            s.send_message("test", json.dumps([i]), expiration=900)
        time.sleep(1)
        receivedlist = []
        msg = r.receive_message()
        logging.warning(msg)
        while msg:
            o = json.loads(msg[0])
            receivedlist.append(o[0])
            msg = r.receive_message()

        logging.warning(receivedlist)
        # Should not receive any messages since we waited for the messages
        # to expire
        self.assertEqual(0, len(receivedlist))


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



from lbmessaging.exchanges.Common import get_connection, check_channel

ALL_EXCHANGES = {}
SERVICES_CHANNELS = {}


def MetaWrapper(exchange_class, exchange_map, service_name):
    """
    Method to dynamically instantiate an exchange and called its methods when
    a mapped method in the service is called. This Wrapper generates the
    associated execution code at runtime
    :param exchange_class: the name of the exchange class
    :param exchange_map: the name of the mapped method in the exchange
    :param service_name: the name of the service
    :return:
    """
    def function(*args, **kwargs):
        exchange = exchange_class(SERVICES_CHANNELS.get(service_name))
        try:
            obj_class = args[0].__class__.__name__
            if obj_class == service_name:
                args = args[1:]
        except:
            pass

        return getattr(exchange, exchange_map)(*args, **kwargs)
    return function


class ServiceMeta(type):
    """
    Meta class used by the BaseService in order to dynamically map the exchanges
    the the services at runtime. Multiple exchanges can be used in the same
    service.
    """
    def __new__(cls, name, bases, dct):
        """
        :param name: the name of the service
        :param bases: the base class of the service
        :param dct: the internal representation of the methods and arguments in
                    the service class
        :return:
        """
        if ALL_EXCHANGES.get(name, None):
            for exchange_data in ALL_EXCHANGES.get(name):
                exchange = exchange_data['exchange']
                service_map = exchange_data['mapping']
                if service_map:
                    for function in service_map.keys():
                        exchange_map = service_map[function]
                        dct[function] = MetaWrapper(exchange, exchange_map,
                                                    name)
        return type.__new__(cls, name, bases, dct)

"""
Parent of BaseService used as artifact to allow python 2 and python 3
compatibility and to prevent duplicated code
"""
try:
    exec('class BaseMetaService(metaclass=ServiceMeta): pass')
except SyntaxError:
    class BaseMetaService(object):
        __metaclass__ = ServiceMeta


class BaseService(BaseMetaService):
    """
    Base class for services that created the internal connection object to the
    message broker based on the init arguments
    """
    def __init__(self, host=None, username=None, passwd=None, port=5671,
                 vhost='/lhcb', use_ssl=True):
        """
        :param host: the RabbitMQ server hostname
        :param username: the RabbitMQ server username
        :param passwd: the RabbitMQ server password
        :param port: the RabbitMQ server port
        :param vhost: the RabbitMQ virtual host that should be used in the
                      connection
        :param use_ssl: if True, an encrypted connection is created, plain text
                        otherwise
        """
        connection = get_connection(host=host, username=username,
                                    passwd=passwd, port=port,
                                    vhost=vhost, use_ssl=use_ssl)
        self.channel = check_channel(None, connection=connection)
        SERVICES_CHANNELS[self.__class__.__name__] = self.channel


class ExposedIn():
    """
    Class used to decorate exchanges in order for the library to declare them
    dynamically in the services.
    """

    def __init__(self, service_name, mapping=None):
        """
        :param service_name: the name of the service in which the decorated
                             exchange should be declared
        :param mapping: a dictionary (of strings) that maps exchange methods
                        to service methods.
        """
        global ALL_EXCHANGES
        if not ALL_EXCHANGES.get(service_name, None):
            ALL_EXCHANGES[service_name] = []
        self.service_exchanges = ALL_EXCHANGES[service_name]
        self.mapping = mapping

    def __call__(self, exchange):
        """
        :param exchange: the decorated exchange
        :return: the original exchange
        """
        if exchange not in self.service_exchanges:
            self.service_exchanges.append({'exchange': exchange,
                                           'mapping': self.mapping})
        return exchange



###############################################################################
# (c) Copyright 2017 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Module in charge of dealing with sending and receiving commands for the
/cvmfs/lhcbdev.cern.ch
'''

import datetime
from collections import namedtuple

from lbmessaging.exchanges.Common import Sender, Receiver, Message, \
    check_channel

__author__ = 'Ben Couturier <ben.couturier@cern.ch>'


# Utility class to ease the use of the messages
CommandMessage = namedtuple('CommandBody',
                            ['command', 'arguments', 'trigger_date'])


def _message_to_command(message):
    """ Trivial utility that converts the command from JSON
    :param message: the message that needs to be converted to command
    :returns the converted rabbitMQ message to Message object with payload as
             command message
    """
    ret = None
    if message and message.body:
        body = CommandMessage(*message.body)
        ret = Message(body, message.method_frame, message.header_frame)
    return ret


def command_retry_count(command):
    """ returns the retry_count kept in the headers
    :param command: the command message whose retries needs to be counted
    :returns the value of the retry_count from the header
    """
    return command.header_frame.headers.get('max_retry', None)


class CommandExchange(object):
    """
    Class in charge of dealing with messages to and from the CvmfsDevExchange
    """

    SUFFIX_RETRY = 'retry'
    SUFFIX_ERROR = 'error'

    def __init__(self, channel, exchange, connection=None):
        """
        Specify the exchange to be used
        :param channel: a pikaBlockingConnection channel that should be used in
                        the broker
        :param exchange: the main exchange name used in the broker

        :param error_exchange: the error exchange that should be used to push
                               a command that failed i.e. the retry_count == 0
        :param connection: a pikaBlockingConnection that should be used in
                        the broker
        :param retry_exchange: the exchange use to retry a message sent
        """
        self._channel = check_channel(channel, connection)
        self._exchange = exchange

    def send_command(self, command, arguments, priority=10, max_retry=3,
                     expiration=None):
        """ Formats and sends the command as JSON
        :param command: the command that should be executed on a CVMFS instance
        :param arguments: the command arguments
        :param priority: the priority of the message
        :param max_retry: the max of retries in case of command failure on CVMFS
        :param expiration: the TTL of the message
        """
        now = datetime.datetime.utcnow().isoformat()
        s = Sender(self._exchange, channel=self._channel)
        s.send_message(command, (command, arguments, now), priority,
                       expiration=expiration,
                       max_retry=max_retry)

    def receive_command(self, queue_name, exchange=None):
        """ Receive one message from the queue
        :param queue_name: the queue on which the broker listens for messages
        :param exchange: the exchange name used in the broker
        :returns the last message converted to Message object with payload as
                 command message
        """
        if exchange is None:
            exchange = self._exchange
        r = Receiver(exchange, queue_name,
                     channel=self._channel)
        message = r.receive_message()
        return _message_to_command(message)

    def receive_command_error(self, queue_name):
        """ Receive one error message from the queue
        :param queue_name: the queue on which the broker listens for messages
        :returns the last message converted to Message object with payload as
                 command message
        """
        error_exchange = '%s.%s' % (queue_name, self.SUFFIX_ERROR)
        error_queue = '%s%s' % (queue_name, self.SUFFIX_ERROR)
        return self.receive_command(error_queue, exchange=error_exchange)

    def receive_all(self, queue_name, exchange=None):
        """ Receive all messages from the queue
        :param queue_name: the queue on which the broker listens for messages
        :param exchange: the exchange name used in the broker
        :returns a list of all the messages converted to Message object with
                 payload as command message
        """
        if exchange is None:
            exchange = self._exchange
        r = Receiver(exchange, queue_name,
                     channel=self._channel)
        msglist = r.receive_messages()
        return [_message_to_command(m) for m in msglist]

    def receive_command_error_all(self, queue_name):
        """ Receive all error messages from the queue
        :param queue_name: the queue on which the broker listens for messages
        :returns the last message converted to Message object with payload as
                 command message
        """
        error_exchange = '%s.%s' % (queue_name, self.SUFFIX_ERROR)
        error_queue = '%s%s' % (queue_name, self.SUFFIX_ERROR)
        return self.receive_all(error_queue, exchange=error_exchange)

    def handle_processing_error(self, command, queue_name=None):
        """ Handles the processing error, i.e. resend if retry_count > 0,
        forward to the error queue otherwise
        :param queue_name: the queue on which the broker listens for messages
        :param command: the command that should be executed on a CVMFS instance
        """

        # In order to keep the retrocompatibility and have the code
        # still working. In the past, the signature was (self, command)
        # without queue_name. If an old client calls this function will
        #  end with an exception concerning the number of arguments.
        # Most probably this exception will be try...catched and logged.
        # This particular exception will be more explicit for the developer
        # in the logs.

        if queue_name is None:
            raise Exception("Please specify the queue name on which the "
                            "receiver listens on")
        # Create first the queues and the exchanges

        # Retry is attached to the same queue_name
        retry_exchange = '%s.%s' % (queue_name, self.SUFFIX_RETRY)
        _ = Receiver(retry_exchange, queue_name, channel=self._channel)

        # Error is attached to the a particular error queue
        error_exchange = '%s.%s' % (queue_name, self.SUFFIX_ERROR)
        error_queue = '%s%s' % (queue_name, self.SUFFIX_ERROR)
        _ = Receiver(error_exchange, error_queue, channel=self._channel)

        retry_count = command_retry_count(command)
        if retry_count > 0:
            s = Sender(retry_exchange, channel=self._channel)
            s.resend_message(command)
        else:
            s = Sender(error_exchange, channel=self._channel)
            s.resend_message(command, keep_retry_count=True)



import logging
import sys
import unittest

from lbmessaging.services.Common import ALL_EXCHANGES
from lbmessaging.services.Common import ExposedIn


@ExposedIn("DumbService_expose")
@ExposedIn("DumbService_expose2")
@ExposedIn("DumbService_expose3")
class DumbExchange():
    pass


@ExposedIn("DumbService_expose2")
class DumbExchange2():
    pass


@ExposedIn("DumbService_expose3", mapping={
    'toto': 'fifi'
})
class DumbExchange3():

    def fifi(self):
        pass


class TestExposeIn(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_services(self):
        self.assertTrue('DumbService_expose' in ALL_EXCHANGES.keys())
        self.assertTrue('DumbService_expose2' in ALL_EXCHANGES.keys())
        self.assertTrue('DumbService_expose3' in ALL_EXCHANGES.keys())

    def test_dumb_service_1(self):
        self.assertEqual(len(ALL_EXCHANGES['DumbService_expose']), 1)
        self.assertEqual(
            type(ALL_EXCHANGES['DumbService_expose'][0]['exchange']),
            type(DumbExchange))

    def test_dumb_service_2(self):
        self.assertEqual(len(ALL_EXCHANGES['DumbService_expose2']), 2)

        self.assertTrue(type(DumbExchange) in [type(x['exchange']) for x in
                                               ALL_EXCHANGES[
                                                   'DumbService_expose2']])
        self.assertTrue(type(DumbExchange2) in [type(x['exchange']) for x in
                                                ALL_EXCHANGES[
                                                    'DumbService_expose2']])

    def test_dumb_service_3(self):
        self.assertEqual(len(ALL_EXCHANGES['DumbService_expose3']), 2)

        self.assertEqual(
            list(ALL_EXCHANGES['DumbService_expose3'][1]['mapping'].keys()),
            ['toto'])

        self.assertTrue(type(DumbExchange) in [type(x['exchange']) for x in
                                               ALL_EXCHANGES[
                                                   'DumbService_expose3']])
        self.assertTrue(type(DumbExchange3) in [type(x['exchange']) for x in
                                                ALL_EXCHANGES[
                                                    'DumbService_expose3']])

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



###############################################################################
# (c) Copyright 2017 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Module in charge of dealing with sending and receiving commands for the
/cvmfs/lhcbdev.cern.ch
'''

from lbmessaging.exchanges.CommandExchange import CommandExchange
from lbmessaging.services.Common import ExposedIn
import os

__author__ = 'Ben Couturier <ben.couturier@cern.ch>'

CVMFSPROD_EXCHANGE = "cvmfsprod.action"


@ExposedIn('CVMFSNightliesService', mapping={
        'prod_send': 'send_command',
        'prod_receive': 'receive_command'
    })
class CvmfsProdExchange(CommandExchange):
    """ Customized exchange for the cvmfsdev startum-0 server """

    def __init__(self, channel):
        """ Constructor that fixes the names of the exchanges
        :param channel: a pikaBlockingConnection channel that should be used in
                        the broker
        """
        super(CvmfsProdExchange, self).__init__(channel,
                                                CVMFSPROD_EXCHANGE)




import logging
import sys
import unittest

from lbmessaging.exchanges.test import get_test_connection

from lbmessaging.exchanges.ContinuousIntegrationExchange import \
    ContinuousIntegrationExchange

QUEUE_NAME = "test_build_ready_queue"
QUEUE_NAME_ENV_KIT = "test_env_kit_ready_queue"


class TestContinuousIntegrationExchange(unittest.TestCase):

    def setUp(self):
        self._connection = get_test_connection()
        self._channel = self._connection.channel()

    def tearDown(self):
        if self._connection:
            self._connection.close()

    def test_build_ready_simple(self):
        """ Check whether a simple command gets through """
        slot = "slot1"
        build_id = "1"
        platform = "x86_64-centos7-gcc62-opt"
        project = "project1"
        broker = ContinuousIntegrationExchange(self._channel)
        broker.receive_all_build_ready(QUEUE_NAME)
        broker.send_build_ready(slot, build_id, platform, project)
        tmp = broker.receive_build_ready(QUEUE_NAME)
        self.assertEqual(slot, tmp.body.slot)
        self.assertEqual(build_id, tmp.body.build_id)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(project, tmp.body.project)

    def test_env_kit_ready_simple(self):
        """ Check whether a simple command gets through """
        flavour = 'dev'
        platform = "centos7"
        version = '243535.2018-01-01'
        broker = ContinuousIntegrationExchange(self._channel)
        broker.receive_all_envkit_ready(QUEUE_NAME_ENV_KIT)
        broker.send_envkit_ready(flavour, platform, version)
        tmp = broker.receive_envkit_ready(QUEUE_NAME_ENV_KIT)
        self.assertEqual(flavour, tmp.body.flavour)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(version, tmp.body.version)
        self.assertEqual(None, tmp.body.url)

    def test_first_come_first_served_build_ready(self):
        """ Check the priority system """

        broker = ContinuousIntegrationExchange(self._channel)

        # purge the queue
        broker.receive_all_build_ready(QUEUE_NAME)

        # Now adding our slots
        slot = "slot1"
        build_id = "1"
        platform = "x86_64-centos7-gcc62-opt"
        project1 = "project1"
        project2 = "project2"
        project3 = "project3"
        broker.send_build_ready(slot, build_id, platform, project1)
        broker.send_build_ready(slot, build_id, platform, project2)
        broker.send_build_ready(slot, build_id, platform, project3)

        # Make sure we get the first come first served
        tmp = broker.receive_build_ready(QUEUE_NAME)
        self.assertEqual(slot, tmp.body.slot)
        self.assertEqual(build_id, tmp.body.build_id)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(project1, tmp.body.project)

        tmp = broker.receive_build_ready(QUEUE_NAME)
        self.assertEqual(slot, tmp.body.slot)
        self.assertEqual(build_id, tmp.body.build_id)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(project2, tmp.body.project)

        tmp = broker.receive_build_ready(QUEUE_NAME)
        self.assertEqual(slot, tmp.body.slot)
        self.assertEqual(build_id, tmp.body.build_id)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(project3, tmp.body.project)

    def test_first_come_first_served_env_kit_ready(self):
        """ Check the priority system """

        broker = ContinuousIntegrationExchange(self._channel)

        # purge the queue
        broker.receive_all_envkit_ready(QUEUE_NAME_ENV_KIT)

        # Now adding our slots
        flavour = 'dev'
        platform = "centos7"
        version1 = '243535.2018-01-01'
        version3 = "243536.2018-01-01"
        version2 = "243537.2018-01-01"

        broker.send_envkit_ready(flavour, platform, version1)
        broker.send_envkit_ready(flavour, platform, version2)
        broker.send_envkit_ready(flavour, platform, version3)

        # Make sure we get the first come first served
        tmp = broker.receive_envkit_ready(QUEUE_NAME_ENV_KIT)
        self.assertEqual(flavour, tmp.body.flavour)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(version1, tmp.body.version)

        tmp = broker.receive_envkit_ready(QUEUE_NAME_ENV_KIT)
        self.assertEqual(flavour, tmp.body.flavour)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(version2, tmp.body.version)

        tmp = broker.receive_envkit_ready(QUEUE_NAME_ENV_KIT)
        self.assertEqual(flavour, tmp.body.flavour)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(version3, tmp.body.version)

    def test_consumer_build_ready(self):
        """ Check whether a simple command gets through """
        slot = "slot1"
        build_id = "1"
        platform = "x86_64-centos7-gcc62-opt"
        project = "project1"
        broker = ContinuousIntegrationExchange(self._channel)
        broker.receive_all_build_ready(QUEUE_NAME)
        broker.send_build_ready(slot, build_id, platform, project)
        broker.send_build_ready("EXIT", build_id, platform, project)

        def checkbuildisok(tmp):
            if tmp.body.slot == "EXIT":
                raise StopIteration
            else:
                self.assertEqual(slot, tmp.body.slot)
                self.assertEqual(build_id, tmp.body.build_id)
                self.assertEqual(platform, tmp.body.platform)
                self.assertEqual(project, tmp.body.project)
        broker.consume_build_ready(checkbuildisok, QUEUE_NAME)

    def test_consumer_env_kit_ready(self):
        """ Check whether a simple command gets through """
        flavour = 'dev'
        platform = "centos7"
        version = '243535.2018-01-01'
        broker = ContinuousIntegrationExchange(self._channel)
        broker.receive_all_envkit_ready(QUEUE_NAME_ENV_KIT)
        broker.send_envkit_ready(flavour, platform, version)
        broker.send_envkit_ready('EXIT', platform, version)

        def checkbuildisok(tmp):
            if tmp.body.flavour == "EXIT":
                raise StopIteration
            else:
                self.assertEqual(flavour, tmp.body.flavour)
                self.assertEqual(platform, tmp.body.platform)
                self.assertEqual(version, tmp.body.version)
        broker.consume_envkit_ready(checkbuildisok, QUEUE_NAME_ENV_KIT)

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()



import logging
import sys
import unittest

from lbmessaging.services.LbCIService import LbCIService
from lbmessaging.exchanges.ContinuousIntegrationExchange \
    import ContinuousIntegrationExchange
from lbmessaging.exchanges.test import get_test_connection


QUEUE_NAME = "test_build_ready_queue"
QUEUE_NAME_ENV_KIT = "test_env_kit_ready_queue"


class TestLbCIService(unittest.TestCase):

    def setUp(self):
        channel = get_test_connection().channel()
        broker = ContinuousIntegrationExchange(channel)
        broker.receive_all_build_ready(QUEUE_NAME)
        broker.receive_all_envkit_ready(QUEUE_NAME_ENV_KIT)

    def tearDown(self):
        pass

    def test_compute_url(self):
        flavour = 'dev'
        platform = "centos7"
        version = '243535.2018-01-01'
        self.assertEqual(
            "http://lhcb-rpm.web.cern.ch/lhcb-rpm/lbenv-kits/"
            "lbenv-kit.dev.centos7.243535.2018-01-01.tar.bz2",
            LbCIService.getEnvKitUrl(flavour, platform, version)
        )
        self.assertEqual(
            "http://toto.cern.ch/here",
            LbCIService.getEnvKitUrl(flavour, platform, version,
                                     url='http://toto.cern.ch/here')
        )

    def test_getEnvKitPrefix(self):
        flavour = 'dev'
        platform = "centos7"
        self.assertEqual(
            "/cvmfs/lhcbdev.cern.ch/new-msr/var/lib/LbEnv/dev/centos7/",
            LbCIService.getEnvKitPrefix(flavour, platform)
        )
        try:
            flavour = '../../../toto'
            LbCIService.getEnvKitPrefix(flavour, platform)
            self.fail("A value error should have been raise")
        except ValueError:
            pass

    def test_services(self):
        srv = LbCIService(vhost='/lhcb-test')
        slot = "slot1"
        build_id = "1"
        platform = "x86_64-centos7-gcc62-opt"
        project = "project1"
        srv.send_build_ready(slot, build_id, platform, project)
        tmp = srv.receive_build_ready(QUEUE_NAME)
        self.assertEqual(slot, tmp.body.slot)
        self.assertEqual(build_id, tmp.body.build_id)
        self.assertEqual(platform, tmp.body.platform)
        self.assertEqual(project, tmp.body.project)

    def test_service_tar_extract(self):
        srv = LbCIService(vhost='/lhcb-test')
        srv.send_envkit_ready('flavour', 'platform', 'version')
        tmp = srv.receive_envkit_ready(QUEUE_NAME_ENV_KIT)
        self.assertEqual('flavour', tmp.body.flavour)
        self.assertEqual('platform', tmp.body.platform)
        self.assertEqual('version', tmp.body.version)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger().setLevel(logging.WARNING)
    unittest.main()
